# Copyright 2019 Brigham Young University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
import sys
import time
from datetime import datetime, timedelta

import jwt
import requests

from .exceptions import JWTHandlerError, JWTVerifyError

JWT_HEADER = 'X-JWT-Assertion'
BYU_JWT_HEADER_ORIGINAL = 'X-JWT-Assertion-Original'


def extract_x5t_from_jwt(_jwt):
    return jwt.get_unverified_header(_jwt)['x5t']


class JWT_Handler(object):
    """BYU_JWT Wraps the process of verifing the BYU JWT"""

    def __init__(self, base_url="https://api.byu.edu"):
        self.base_url = base_url
        self.jwks_data = {'pubkeys': {}, 'issuer': {}, 'ttl': 0}
        self.request_headers = {
            'User-Agent': 'BYU-JWT-Python-SDK/2.0 (Python {})'.format(sys.version.replace('\n', ''))}

    def get_signing_cert(self):
        """
        Returns public jwks cert data from cache or fetches if needed
        """
        if time.time() > self.jwks_data['ttl']:
            well_known = self._get_wellknown_data()
            self.jwks_data['issuer'] = well_known['issuer']
            jwks, ttl = self._get_jwks_data(well_known['jwks_uri'])
            self.jwks_data['pubkeys'] = self._extract_public_keys(jwks)
            self.jwks_data['ttl'] = ttl
        return self.jwks_data['pubkeys']

    def _extract_public_keys(self, jwks):
        # TODO Need to unit test as this is undocumented feature of pyjwt library
        key_set = {}
        for key in jwks['keys']:
            thumbprint = key['x5t']
            key_set[thumbprint] = jwt.algorithms.RSAAlgorithm.from_jwk(json.dumps(key))
        return key_set

    def _get_wellknown_data(self):
        url = '{}/.well-known/openid-configuration'.format(self.base_url)
        try:
            response = requests.get(url, headers=self.request_headers)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            raise JWTHandlerError("Error getting .well-known data") from e

    def _get_jwks_data(self, jwks_uri):
        try:
            response = requests.get(jwks_uri, headers=self.request_headers)
            response.raise_for_status()
            cache_control = response.headers.get(
                'Cache-Control', 'public max-age=3600').split('=')[1]
            ttl = datetime.now() + timedelta(seconds=int(cache_control))
            return response.json(), int(ttl.timestamp())
        except requests.exceptions.RequestException as e:
            raise JWTHandlerError("Error getting jwks data") from e

    def is_valid(self, jwt_to_validate):
        try:
            self.decode(jwt_to_validate)
            return True
        except Exception:
            return False

    def decode(self, jwt_to_decode, verify=True):
        try:
            x5t = extract_x5t_from_jwt(jwt_to_decode)
            pubkeys = self.get_signing_cert()
            decoded_jwt = jwt.decode(jwt_to_decode,
                                     pubkeys[x5t],
                                     verify=verify,
                                     issuer=self.jwks_data['issuer'],
                                     leeway=2,
                                     algorithms=['RS256'])
            decoded_jwt = self.add_byu_output_structure(decoded_jwt)
            return decoded_jwt
        except jwt.exceptions.PyJWTError as e:
            raise JWTVerifyError("Invalid JWT") from e

    def add_byu_output_structure(self, decoded_jwt):
        decoded_jwt['byu'] = {
            'client': {
                'byuId': decoded_jwt.pop('http://byu.edu/claims/client_byu_id'),
                'claimSource': decoded_jwt.pop('http://byu.edu/claims/client_claim_source'),
                'netId': decoded_jwt.pop('http://byu.edu/claims/client_net_id'),
                'personId': decoded_jwt.pop('http://byu.edu/claims/client_person_id'),
                'preferredFirstName': decoded_jwt.pop('http://byu.edu/claims/client_preferred_first_name'),
                'prefix': decoded_jwt.pop('http://byu.edu/claims/client_name_prefix'),
                'restOfName': decoded_jwt.pop('http://byu.edu/claims/client_rest_of_name'),
                'sortName': decoded_jwt.pop('http://byu.edu/claims/client_sort_name'),
                'subscriberNetId': decoded_jwt.pop('http://byu.edu/claims/client_subscriber_net_id'),
                'suffix': decoded_jwt.pop('http://byu.edu/claims/client_name_suffix'),
                'surname': decoded_jwt.pop('http://byu.edu/claims/client_surname'),
                'surnamePosition': decoded_jwt.pop('http://byu.edu/claims/client_surname_position')
            }
        }
        decoded_jwt['wso2'] = {
            'apiContext': decoded_jwt.pop('http://wso2.org/claims/apicontext'),
            'application': {
                'id': decoded_jwt.pop('http://wso2.org/claims/applicationid'),
                'name': decoded_jwt.pop('http://wso2.org/claims/applicationname'),
                'tier': decoded_jwt.pop('http://wso2.org/claims/applicationtier'),
            },
            'clientId': decoded_jwt.pop('http://wso2.org/claims/client_id'),
            'endUser': decoded_jwt.pop('http://wso2.org/claims/enduser'),
            'endUserTenantId': decoded_jwt.pop('http://wso2.org/claims/enduserTenantId'),
            'keyType': decoded_jwt.pop('http://wso2.org/claims/keytype'),
            'subscriber': decoded_jwt.pop('http://wso2.org/claims/subscriber'),
            'tier': decoded_jwt.pop('http://wso2.org/claims/tier'),
            'userType': decoded_jwt.pop('http://wso2.org/claims/usertype'),
            'version': decoded_jwt.pop('http://wso2.org/claims/version'),
        }
        if 'http://byu.edu/claims/resourceowner_person_id' in decoded_jwt:
            decoded_jwt['byu']['resourceOwner'] = {
                'byuId': decoded_jwt.pop('http://byu.edu/claims/resourceowner_byu_id'),
                'netId': decoded_jwt.pop('http://byu.edu/claims/resourceowner_net_id'),
                'personId': decoded_jwt.pop('http://byu.edu/claims/resourceowner_person_id'),
                'preferredFirstName': decoded_jwt.pop('http://byu.edu/claims/resourceowner_preferred_first_name'),
                'prefix': decoded_jwt.pop('http://byu.edu/claims/resourceowner_prefix'),
                'restOfName': decoded_jwt.pop('http://byu.edu/claims/resourceowner_rest_of_name'),
                'sortName': decoded_jwt.pop('http://byu.edu/claims/resourceowner_sort_name'),
                'suffix': decoded_jwt.pop('http://byu.edu/claims/resourceowner_suffix'),
                'surname': decoded_jwt.pop('http://byu.edu/claims/resourceowner_surname'),
                'surnamePosition': decoded_jwt.pop('http://byu.edu/claims/resourceowner_surname_position'),
            }
        return decoded_jwt



class BYUJWTError(Exception):
    """Error handling BYU JWT"""


class JWTVerifyError(BYUJWTError):
    """Error Verifing JWT"""


class JWTHandlerError(BYUJWTError):
    """Error while handling JWT"""



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['byu_jwt']

package_data = \
{'': ['*']}

install_requires = \
['cryptography>=2.4,<3.0', 'pyjwt>=1.7,<2.0', 'requests>=2.21,<3.0']

setup_kwargs = {
    'name': 'byu-jwt',
    'version': '2.0.0',
    'description': '',
    'long_description': '# byu-jwt-python\nA python JWT validator that does all the BYU specific stuff as well as handle caching well-known and cert fetching\n\n# Installation\n`pip install byu_jwt`\n\n## API\n\n---\n**Note: It is important to declare the handler at a global level. This allows the caching of the well-known data as well as using the cache-control headers on the certificates only re-fetching those when cache-control has timed out. Reinitializing the class object will negate any benefit of the caching**\n\n---\nInstantiate the class and reuse the object to utilize caching:\n```python\nimport byu_jwt\nbyujwt = byu_jwt.JWT_Handler()\n```\n\n### Check only if JWT is valid\n```python\nassert byujwt.is_valid(jwt_to_validate)\n```\n\n### Decode JWT and Check validity\n```python\ntry:\n    jwt = byujwt.decode(jwt_to_validate)\n    return f"Hello, {jwt[\'preferredFirstName\']}"\nexcept byu_jwt.exceptions.JWTVerifyError as ex_info:\n    return "Invalid JWT"\nexcept byu_jwt.exceptions.JWTHandlerError as ex_info:\n    return "Error attempting to verify the jwt"\n```\n\n### JWT Header Names\n\nBYU\'s API Manager creates an HTTP header that contains a signed [JWT](https://jwt.io/). The names of the designed BYU signed headers can be referenced here for lookup convenience.\n\n### BYU_JWT_HEADER_CURRENT\n\nThe property containing the name of the HTTP header that contains the BYU signed JWT sent directly from BYU\'s API Manager.\n\nValue is X-JWT-Assertion.\n\nExample\n\n```python\ncurrent_jwt_header = byu_jwt.JWT_HEADER\n```\n\n### BYU_JWT_HEADER_ORIGINAL\n\nThe property containing the name of the HTTP header that contains the BYU signed JWT forwarded on from a service that received the BYU signed JWT sent directly from BYU\'s API Manager.\n\nValue is X-JWT-Assertion-Original.\n\nExample\n\n```python\noriginal_jwt_header = byu_jwt.JWT_HEADER_ORIGINAL\n```\n\n### Example Python Lambda function that makes use of caching\n```python\nimport byu_jwt\n\nbyujwt = byu_jwt.JWT_Handler()\n\ndef handler(event, context):\n    jwt_to_decode = event[\'headers\'][byu_jwt.JWT_HEADER]\n    try:\n        jwt = byujwt.decode(jwt_to_validate)\n        return {\'statusCode\': 200, \'body\': f\'Hello, {jwt["preferredFirstName"]}\'}\n    except byu_jwt.exceptions.JWTVerifyError as ex_info:\n        return {\'statusCode\': 403, \'body\': "Invalid JWT"}\n    except byu_jwt.exceptions.JWTHandlerError as ex_info:\n        return {\'statusCode\': 500, \'body\': "Error attempting to verify the jwt"}\n```\n\n\n### Example Decoded JWT Structure\n```json\n{\n  "iss": "https://api.byu.edu",\n  "exp": 1545425710,\n  "byu": {\n    "client": {\n      "byuId": "",\n      "claimSource": "",\n      "netId": "",\n      "personId": "",\n      "preferredFirstName": "",\n      "prefix": "",\n      "restOfName": "",\n      "sortName": "",\n      "subscriberNetId": "",\n      "suffix": "",\n      "surname": "",\n      "surnamePosition": ""\n    },\n    "resourceOwner": {\n      "byuId": "",\n      "netId": "",\n      "personId": "",\n      "preferredFirstName": "",\n      "prefix": "",\n      "restOfName": "",\n      "sortName": "",\n      "suffix": "",\n      "surname": "",\n      "surnamePosition": ""\n    }\n  },\n  "wso2": {\n    "apiContext": "",\n    "application": {\n      "id": "",\n      "name": "",\n      "tier": ""\n    },\n    "clientId": "",\n    "endUser": "",\n    "endUserTenantId": "",\n    "keyType": "",\n    "subscriber": "",\n    "tier": "",\n    "userType": "",\n    "version": ""\n  }\n}\n```',
    'author': 'Nate Peterson',
    'author_email': 'nate_peterson@byu.edu',
    'url': 'https://github.com/byu-oit/byu-jwt-python',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.4,<4.0',
}


setup(**setup_kwargs)
